# Copyright 2016 Stephan Rave. All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this list of conditions and the following
#   disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following
#   disclaimer in the documentation and/or other materials provided with the distribution.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import numpy as np

from pymor.discretizations.basic import DiscretizationBase
from pymor.vectorarrays.interfaces import VectorArrayInterface
from pymor.vectorarrays.numpy import NumpyVectorArray, NumpyVectorSpace
from pymor.operators.interfaces import OperatorInterface
from pymor.operators.constructions import VectorOperator
from pymor.operators.basic import OperatorBase
from pymor.parameters.base import Parameter

from group import R1ShiftGroupAction, R2ShiftGroupAction, OrthogonalityPhaseCondition


def freeze(d, d_data):
    g = R1ShiftGroupAction(d_data['grid']) if d_data['grid'].dim == 1 else R2ShiftGroupAction(d_data['grid'])
    phase_condition = OrthogonalityPhaseCondition(g, d.operator)
    d_f = FrozenDiscretization(initial_data=d.initial_data, operator=d.operator, group_action=g,
                               phase_condition=phase_condition, products=d.products, parameter_space=d.parameter_space,
                               T=d.T, nt=d.time_stepper.nt, visualizer=d.visualizer)
    return d_f


class FrozenDiscretization(DiscretizationBase):

    def __init__(self, T, nt, initial_data, operator, group_action, phase_condition, products=None,
                 frozen_operator=None, num_values=None, parameter_space=None, visualizer=None,
                 cache_region=None, name=None):
        assert isinstance(initial_data, (VectorArrayInterface, OperatorInterface))
        assert not isinstance(initial_data, OperatorInterface) or initial_data.source == NumpyVectorSpace(1)
        assert isinstance(operator, OperatorInterface)
        if isinstance(initial_data, VectorArrayInterface):
            initial_data = VectorOperator(initial_data, name='initial_data')
        assert operator.source == operator.range == initial_data.range
        assert '_t' not in operator.parameter_type

        if not frozen_operator:
            assert 'numerical_flux_flux' in operator.with_arguments
            frozen_flux = operator.numerical_flux.flux + group_action.lie_algebra_flux
            if not hasattr(operator.numerical_flux, 'flux_derivative'):
                frozen_operator = operator.with_(numerical_flux_flux=frozen_flux)
            else:
                frozen_flux_derivative = operator.numerical_flux.flux_derivative + group_action.lie_algebra_flux_derivative
                frozen_operator = operator.with_(numerical_flux_flux=frozen_flux,
                                                 numerical_flux_flux_derivative=frozen_flux_derivative)
        operators = {'operator': operator, 'frozen_operator': frozen_operator}
        functionals = {}
        vector_operators = {'initial_data': initial_data}
        super(FrozenDiscretization, self).__init__(operators=operators, functionals=functionals,
                                                   vector_operators=vector_operators,
                                                   products=products, visualizer=visualizer,
                                                   cache_region=cache_region, name=name)
        self.T = T
        self.nt = nt
        self.solution_space = operator.source
        self.initial_data = initial_data
        self.operator = operator
        self.frozen_operator = frozen_operator
        self.very_frozen_operator = VeryFrozenOperator(frozen_operator, phase_condition)
        self.group_action = group_action
        self.phase_condition = phase_condition
        self.num_values = num_values
        self.build_parameter_type(inherits=(initial_data, operator))
        self.parameter_space = parameter_space

    def with_(self, **kwargs):
        assert set(kwargs.keys()) <= self.with_arguments
        assert 'operators' not in kwargs or kwargs['operators'].viewkeys() <= set(('operator', 'frozen_operator'))
        assert 'functionals' not in kwargs or kwargs['functionals'].viewkeys() <= set(tuple())
        assert 'vector_operators' not in kwargs or kwargs['vector_operators'].viewkeys() <= set(('initial_data',))
        assert 'operators' not in kwargs or not set(kwargs['operators']).intersection(kwargs.viewkeys())
        assert 'functionals' not in kwargs or not set(kwargs['functionals']).intersection(kwargs.viewkeys())
        assert 'vector_operators' not in kwargs or not set(kwargs['vector_operators']).intersection(kwargs.viewkeys())
        if 'operators' in kwargs:
            kwargs.update(kwargs.pop('operators'))
        if 'functionals' in kwargs:
            kwargs.update(kwargs.pop('functionals'))
        if 'vector_operators' in kwargs:
            kwargs.update(kwargs.pop('vector_operators'))

        return super(FrozenDiscretization, self).with_(**kwargs)

    def _solve(self, mu=None):
        mu = self.parse_parameter(mu)
        if mu is None:
            mu = Parameter()
        else:
            mu = mu.copy()

        if not self.logging_disabled:
            self.logger.info('Solving {} for {} ...'.format(self.name, mu))

        U0 = self.initial_data.as_vector(mu)
        operator = self.operator
        frozen_operator = self.frozen_operator
        dt = self.T / self.nt
        num_values = self.num_values or (self.nt + 1)
        R = operator.source.empty(reserve=num_values)
        R.append(U0)

        U = U0.copy()
        etas = np.zeros((num_values,) + self.frozen_operator.parameter_type['_group'], dtype=np.float)
        gs = np.zeros_like(etas, dtype=np.float)
        g = np.zeros(self.frozen_operator.parameter_type['_group'], dtype=np.float)
        phase_condition = self.phase_condition

        for n in range(1, self.nt + 1):
            eta = phase_condition.solve(U, mu)
            g += eta * dt
            mu['_group'] = eta
            U.axpy(-dt, frozen_operator.apply(U, mu=mu))
            if n * (num_values / self.nt) > len(R):
                ind = len(R)
                R.append(U)
                etas[ind - 1] = eta
                gs[ind] = g

        R.etas = etas
        R.group = gs

        return R

    def reconstruct(self, U):
        R = U.empty(U.dim, reserve=len(U))
        for i, g in enumerate(U.group):
            R.append(self.group_action.apply(g, U.copy(ind=i)))
        return R


class VeryFrozenOperator(OperatorBase):

    def __init__(self, frozen_operator, phase_condition):
        self.source = frozen_operator.source
        self.range = frozen_operator.range
        self.frozen_operator = frozen_operator
        self.phase_condition = phase_condition
        pt = dict(frozen_operator.parameter_type)
        del pt['_group']
        self.build_parameter_type(pt, local_global=True)

    def apply(self, U, ind=None, mu=None):
        mu = self.parse_parameter(mu).copy()
        assert isinstance(U, NumpyVectorArray)
        assert U in self.source
        ind = range(len(U)) if ind is None else ind
        R = self.frozen_operator.range.empty(reserve=len(ind))
        for i in ind:
            u = U.copy(ind=i)
            eta = self.phase_condition.solve(u, mu)
            mu['_group'] = eta
            R.append(self.frozen_operator.apply(u, mu=mu))
        return R
