# Copyright 2016 Stephan Rave. All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this list of conditions and the following
#   disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following
#   disclaimer in the documentation and/or other materials provided with the distribution.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from __future__ import absolute_import, division, print_function

import math as m

import numpy as np

from pymor.grids.rect import RectGrid
from pymor.grids.oned import OnedGrid
from pymor.vectorarrays.numpy import NumpyVectorArray
from pymor.core.interfaces import ImmutableInterface
from pymor.functions.basic import GenericFunction, ConstantFunction
from pymor.operators.fv import LinearAdvectionLaxFriedrichs
from pymor.operators.numpy import NumpyMatrixOperator
from pymor.grids.boundaryinfos import EmptyBoundaryInfo
from pymor.parameters.base import Parametric


class R2ShiftGroupAction(ImmutableInterface):

    def __init__(self, grid):
        assert isinstance(grid, RectGrid)
        assert len(grid.boundaries(0)) == 0
        self.grid = grid

        def flux_function(U, mu):
            v = - mu['_group'].reshape((1,) * (U.ndim-1) + (2,))
            return U * v
        flux = GenericFunction(flux_function, dim_domain=1, shape_range=(2,),
                               parameter_type={'_group': (2,)}, name='R2ShiftFlux')
        self.lie_algebra_flux = flux

        def flux_function_derivative(U, mu):
            return np.ones(U.shape[:-1] + (2,)) * (-mu['_group'])
        flux_derivative = GenericFunction(flux_function_derivative, dim_domain=1, shape_range=(2,),
                                          parameter_type={'_group': (2,)}, name='R2ShiftFluxDerivative')
        self.lie_algebra_flux_derivative = flux_derivative

        bi = EmptyBoundaryInfo(grid)
        self.lie_algebra_basis = [LinearAdvectionLaxFriedrichs(grid, bi,
                                                               ConstantFunction(np.array([-1., 0.]), dim_domain=2),
                                                               lxf_lambda=np.inf),
                                  LinearAdvectionLaxFriedrichs(grid, bi,
                                                               ConstantFunction(np.array([0., -1.]), dim_domain=2),
                                                               lxf_lambda=np.inf)]

    def apply(self, g, U):
        assert len(U) == 1
        assert g.shape == (2,)

        grid = self.grid
        nx0 = grid.x0_num_intervals
        nx1 = grid.x1_num_intervals
        hx0 = grid.x0_diameter
        hx1 = grid.x1_diameter
        g = -g
        U = U.data.reshape((nx1, nx0))

        I00 = np.repeat(np.array((np.arange(nx0) + (m.floor(g[0] / hx0))) % nx0, dtype=np.int)[np.newaxis, :],
                        nx1, axis=0)
        I01 = np.repeat(np.array((np.arange(nx0) + (m.floor(g[0] / hx0) + 1)) % nx0, dtype=np.int)[np.newaxis, :],
                        nx1, axis=0)
        I10 = np.repeat(np.array((np.arange(nx1) + (m.floor(g[1] / hx1))) % nx1, dtype=np.int)[:, np.newaxis],
                        nx0, axis=1)
        I11 = np.repeat(np.array((np.arange(nx1) + (m.floor(g[1] / hx1) + 1)) % nx1, dtype=np.int)[:, np.newaxis],
                        nx0, axis=1)

        x0_ratio = (g[0] - (m.floor(g[0] / hx0)) * hx0) / hx0
        x1_ratio = (g[1] - (m.floor(g[1] / hx1)) * hx1) / hx1

        R = U[I10, I00] * ((1 - x1_ratio) * (1 - x0_ratio))
        R += U[I11, I00] * ((x1_ratio) * (1 - x0_ratio))
        R += U[I10, I01] * ((1 - x1_ratio) * (x0_ratio))
        R += U[I11, I01] * ((x1_ratio) * (x0_ratio))

        return NumpyVectorArray(R.ravel(), copy=False)


class R1ShiftGroupAction(ImmutableInterface):

    def __init__(self, grid):
        assert isinstance(grid, OnedGrid)
        assert len(grid.boundaries(0)) == 0
        self.grid = grid

        def flux_function(U, mu):
            return U * (-mu['_group'])
        flux = GenericFunction(flux_function, dim_domain=1, shape_range=(1,),
                               parameter_type={'_group': (1,)}, name='R1ShiftFlux')
        self.lie_algebra_flux = flux

        def flux_function_derivative(U, mu):
            return np.ones_like(U) * (-mu['_group'])
        flux_derivative = GenericFunction(flux_function_derivative, dim_domain=1, shape_range=(1,),
                                          parameter_type={'_group': (1,)}, name='R1ShiftFluxDerivative')
        self.lie_algebra_flux_derivative = flux_derivative

        bi = EmptyBoundaryInfo(grid)

        self.lie_algebra_basis = [LinearAdvectionLaxFriedrichs(grid, bi,
                                                               ConstantFunction(np.array([-1.]), dim_domain=1),
                                                               lxf_lambda=np.inf)]

    def apply(self, g, U):
        assert len(U) == 1
        assert g.shape == (1,)

        grid = self.grid
        n = grid.size(0)
        h = grid.volumes(0)[0]
        g = -g
        U = U.data.ravel()

        I0 = np.array((np.arange(n) + (m.floor(g / h))) % n, dtype=np.int)
        I1 = np.array((np.arange(n) + (m.floor(g / h) + 1)) % n, dtype=np.int)

        ratio = (g[0] - (m.floor(g[0] / h)) * h) / h

        R = U[I0] * (1 - ratio)
        R += U[I1] * ratio

        return NumpyVectorArray(R, copy=False)


class OrthogonalityPhaseCondition(ImmutableInterface, Parametric):
    def __init__(self, group_action, operator):
        self.group_action = group_action
        self.operator = operator
        self.build_parameter_type(inherits=(operator,))
        self._lie_algebra_basis = group_action.lie_algebra_basis

    def solve(self, U, mu=None):
        AU = self.operator.apply(U, mu=mu)
        LU = [op.apply(U) for op in self._lie_algebra_basis]
        LHS = np.array([[V.pairwise_dot(W)[0] for W in LU] for V in LU])
        RHS = - np.array([AU.pairwise_dot(V)[0] for V in LU])
        return np.linalg.solve(LHS, RHS)

    def projected(self, RB):
        LRB = [op.apply(RB) for op in self._lie_algebra_basis]
        LHS_data = [[NumpyMatrixOperator(lrb.dot(lrb2)) for lrb2 in LRB] for lrb in LRB]
        RHS_ops = tuple(self.operator.projected(source_basis=RB, range_basis=lrb) for lrb in LRB)
        return ProjectedOrthogonalityPhaseCondition(LHS_data, RHS_ops)


class ProjectedOrthogonalityPhaseCondition(ImmutableInterface, Parametric):
    def __init__(self, LHS_data, RHS_ops):
        self.LHS_data = LHS_data
        self.RHS_ops = RHS_ops
        self.build_parameter_type(inherits=RHS_ops)

    def solve(self, U, mu=None):
        LHS = np.array([[l.pairwise_apply2(U, U)[0] for l in lr] for lr in self.LHS_data])
        RHS = - np.array([op.pairwise_apply2(U, U, mu=mu)[0] for op in self.RHS_ops])
        return np.linalg.pinv(LHS).dot(RHS)
